import torch
from transformers import AutoModel, AutoTokenizer

from .utils import visualize_tokens

#Mean Pooling - Take attention mask into account for correct averaging
def mean_pooling(model_output, attention_mask, device="mps"):
    device = torch.device(device)  # Create a torch.device object for the MPS device
    token_embeddings = model_output.last_hidden_state.to(device)  # Move token_embeddings to MPS device
    attention_mask = attention_mask.to(device)  # Move attention_mask to MPS device
    input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

class EmbeddingModel:
    '''
    EmbeddingModel runs a transformer model and returns the embedding for a given text.
    '''
    def __init__(self, model_name='sentence-transformers/all-MiniLM-L6-v2'):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=True) # use_fast=True

        self.model = AutoModel.from_pretrained(model_name)
        self.dimension = self.model.embeddings.position_embeddings.embedding_dim
        self.max_seq_length = self.model.embeddings.position_embeddings.num_embeddings
        

        print("Tokenizer:", self.tokenizer)
        # print("Dimension:", self.dimension)
        # print("Max sequence length:", self.max_seq_length)

    def embed(self, texts, max_seq_length=256, device="mps"):

        device = torch.device(device)  # Create a torch.device object
        self.model.to(device)  # Move the model to the specified device

        encoded_input = self.tokenizer(texts, padding=True, truncation=True, return_tensors='pt', max_length=max_seq_length)
        encoded_input = {name: tensor.to(device) for name, tensor in encoded_input.items()}  # Move all input tensors to the specified device

        with torch.no_grad():
            model_output = self.model(**encoded_input)

        embeddings = mean_pooling(model_output, encoded_input['attention_mask'], device=device)
        tensor_embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
        np_embeddings = tensor_embeddings.cpu().numpy()  # Move tensor to CPU before converting to numpy

        # Visualize tokens with colors
        # tokens = [self.tokenizer.decode([input_id]) for row in encoded_input['input_ids'] for input_id in row]
        # visualize_tokens(tokens)

        return np_embeddings

    def token_count(self, texts):
        tokens = 0
        for text in texts:
            tokens+=len(self.tokenizer.tokenize(text))
    


import numpy as np
import pysbd
import PyPDF2
import itertools
from transformers import AutoTokenizer, AutoModel
import regex as re

# def chop_and_chunk(text, max_seq_length=256):
#     if isinstance(text, str):
#         if len(text) < max_seq_length*4:
#             return [text]
#         text = [text]
#     segmenter = pysbd.Segmenter(language="en", clean=False)
#     chunks = []
#     for i, sentence in enumerate(text):
#         if len(sentence) < max_seq_length*4:
#             chunks.append(sentence)
#         else:
#             segs = segmenter.segment(sentence)
#             buffer = "" # buffer for the current chunk
#             for seg in segs:
#                 if len(buffer) + len(seg) > max_seq_length*4:
#                     chunks.append(buffer)
#                     buffer = ""
#                 buffer += seg
#                 if len(buffer) > max_seq_length*4:
#                     chunks.append(buffer)
#                     buffer = ""
#                 else:
#                     buffer += ""
#     return chunks

def chop_and_chunk(text, max_seq_length=256):
    """
    Chop and chunk a text into smaller pieces of text. 
    
    Args:
    text: string, list of strings, or array of strings 
    max_seq_length: maximum length of the text
    """
    if isinstance(text, str):
        text = [text]
        
    # If text is already split into chunks by newlines, simply return it 
    if all('\n' in t for t in text):
        return text 
        
    chunks = []
    for t in text: 
        # Split by newlines 
        parts = re.split('\n+', t)  
        
        for p in parts:
            tokens = p.split()
            chunk = ''
            count = 0
            for t in tokens:
                if count + len(t) < max_seq_length:
                    count += len(t) 
                    chunk += t + ' '
                else:
                    chunks.append(chunk.strip())
                    count = 0
                    chunk = ''
            if chunk != '':
                chunks.append(chunk.strip())
    return chunks
    
def cos_sim(a, b):
    sims = a @ b.T
    sims /= np.linalg.norm(a) * np.linalg.norm(b, axis=1) 
    return sims

def load_file(pdf_path):
    extracted_text = []
    with open(pdf_path, "rb") as file:
        reader = PyPDF2.PdfReader(file)
        for page in iter(reader.pages):
            extracted_text.append(page.extract_text())  
    return extracted_text

def visualize_tokens(token_values: list[str]) -> None:
        backgrounds = itertools.cycle(
            ["\u001b[48;5;{}m".format(i) for i in [167, 179, 185, 77, 80, 68, 134]]
        )
        interleaved = itertools.chain.from_iterable(zip(backgrounds, token_values))
        print(("".join(interleaved) + "\u001b[0m"))

def token_count(texts):
        tz = AutoTokenizer.from_pretrained('sentence-transformers/all-MiniLM-L6-v2', use_fast=True)
        tokens = 0
        for text in texts:
            tokens+=len(tz.tokenize(text, padding=True, truncation=True))
        return tokens


import numpy as np
from uuid import uuid4
from .model import EmbeddingModel
from .utils import chop_and_chunk, cos_sim

class VLite:
    '''
    vlite is a simple vector database that stores vectors in a numpy array.
    '''
    def __init__(self, collection='vlite.npz'):
        self.collection = collection
        self.model = EmbeddingModel()
        try:
            with np.load(self.collection, allow_pickle=True) as data:
                self.texts = data['texts'].tolist()
                self.metadata = data['metadata'].tolist()
                self.vectors = data['vectors']
        except FileNotFoundError:
            self.texts = []
            self.metadata = {}
            self.vectors = np.empty((0, self.model.dimension))
    
    def add_vector(self, vector):
        self.vectors = np.vstack((self.vectors, vector))

    def get_similar_vectors(self, vector, top_k=5):
        sims = cos_sim(vector, self.vectors)
        sims = sims[0]
        # print("[get_similar_vectors] Sims:", sims.shape)
        top_k_idx = np.argsort(sims)[::-1][:top_k]
        # print("[get_similar_vectors] Top k idx:", top_k_idx)
        # print("[get_similar_vectors] Top k sims:", sims[top_k_idx])
        return top_k_idx, sims[top_k_idx]

    def memorize(self, text, id=None, metadata=None):
        id = id or str(uuid4())
        chunks = chop_and_chunk(text)
        encoded_data = self.model.embed(texts=chunks, device="mps")
        self.vectors = np.vstack((self.vectors, encoded_data))
        for chunk in chunks:
            self.texts.append(chunk)
            idx = len(self.texts) - 1
            self.metadata[idx] = metadata or {}
            self.metadata[idx]['index'] = id or idx
        self.save()
        return id, self.vectors

    def remember(self, text=None, id=None, top_k=5):
        if id:
            return self.metadata[id]
        if text:

            sims = cos_sim(self.model.embed(texts=text, device="cpu") , self.vectors)
            print("[remember] Sims:", sims.shape)
            sims = sims[0]

            # Use np.argpartition to partially sort only the top 5 values
            top_5_idx = np.argpartition(sims, -top_k)[-top_k:]  

            # Use np.argsort to sort just those top 5 indices
            top_5_idx = top_5_idx[np.argsort(sims[top_5_idx])[::-1]]  

            # print("[remember] Top k sims:", sims[top_5_idx])
            return [self.texts[idx] for idx in top_5_idx], sims[top_5_idx]
            
    def save(self):
        with open(self.collection, 'wb') as f:
            np.savez(f, texts=self.texts, metadata=self.metadata, vectors=self.vectors)



from setuptools import setup, find_packages

setup(
    name='vlite',
    version='0.1.4',
    author='Surya Dantuluri',
    author_email='surya@suryad.com',
    description='A simple vector database that stores vectors in a numpy array.',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pysbd',
        'PyPDF2',
        'transformers',
        'torch',
        'uuid'
    ],
)
